TITLE "APP-NOTE PER I2C-BUS"
LIST P=16C54 , C=132

; PIC16C5X to 24LC01B or 24LC02B EEPROM communication code.
; Based on Franco code.

; FUSE SETTINGS:
; OSC = XT, WDT = ON, CP = OFF, CKSUM =  66F3

; INTRODUCTION:
; The Microchip MTA85XXX family of microntrollers are multichip modules 
; which contain a PIC16C5X microcontroller and a 24LC01B or 24LC02B EEPROM.
; This appliction note is intended to provide users with highly compressed
; assembly code for communication between the EEPROM and the Microcontroller,
; which will leave the user a maximum amount of code space for the core 
; application.

; The code contained in this application note will allow the user to read 
; and write the EEPROM, with acknowledge checking.  This application note 
; will function with both bonding options of the MTA85XXX.  This application
; note will light LEDs on port B to indicate that the data is being written
; to the EEPROM or to output data read from the EEPROM.

; For use in a users application, the user needs to cut the EEPROM 
; subroutines, EEPROM initilization in MAIN, and the variable
; declarations that apply to the EEPROM subroutines.  If the user selects
; from the start of the code to the line which reads "Cut from here up..." 
; all necessary code will be selected.


; EEPROM SUBROUTINES:
; Write_Byte   - Write data supplied in EEDATA at address supplied in EEADD
; Read_Random  - Read data into EEDATA from address supplied in EEADD 
; Read_Current - Read data into EEDATA from address currently in EEPROM


; CODE SIZE:
; EEPROM SUBROUTINES     - 71 bytes
; EEPROM INITILIZATION   -  4 bytes  - Initilize SDA and SCL
; REGISTER INITILIZATION -  4 bytes  - set up EEADD and EEDATA
; READ OR WRITE LOOP  7 to 14 bytes  - approximate number of bytes to
;                                      control EEPROM subroutines
; TOTAL BYTES:       86 to 95 bytes for EEPROM communication including all
;                             subroutines and control logic in MAIN routine
; THIS APPLICATION NOTE  - 148 bytes


; TIMINGS (assume 4MHz):
; Write operation will take approximately 350us + standard write time.
; Read random address operation will take approximately 350us.
; Read current address will take approximately 250us.


; OPERATION:
; The SDA and SCL bits of the EEPROM must be initilized to '1's before
;  calling any of the EEPROM subroutines.  This is done in the beginning
;  of the MAIN routine.

; The registers EEADD and EEDATA are used to communicate EEPROM address and
;  data information to and from the EEPROM.  See the I/O requirements of 
;  each EEPROM subroutine in the commments below to see how these two 
;  registers are utilized in each case.

; During a Write_Byte operation the master will set address and data 
;  information into the EEADD and EEDATA registers.  After receiving the 
;  acknowledge from the EEPROM that the current EEADD and EEDATA have been 
;  received, new values will be loaded into the EEADD and EEDATA registers
;  The EEPROM initiates a self timed write after the acknowledge, during 
;  which time it will not accept or acknowledge other input.  The master 
;  falls into a loop trying to input the new address and data into the 
;  EEPROM.  This loop will continue until the EEPROM finishes the self timed
;  write and acknowledges receipt of the new address and data information.  
;  Note WDT should be cleared during the EE write time.

; A read will attempt to read the data at the supplied or current address
;  until an acknowledge is received.  If an acknowledge is not received, the
;  part will time out after a WDT period.  If a Read_Random is being done
;  the address is supplied in EEADD.  In both Read_Random and Read_Current 
;  the data is read into the EEDATA register.  

; Bits RA0 and RA1 are used to determine if the user would like to read
;  or write the array and to select the array size (1K or 2K).  Bits RA2 and
;  RA3 are used for SCL and SDA.
;    RA0 = '0' for 1K EEPROM             RA1 = '0' for read  EEPROM
;    RA0 = '1' for 2K EEPROM             RA1 = '1' for write EEPROM
;    SCL = RA2                           SDA = RA3, needs 10K pullup

; This application note uses the 8-bit portB to provide an LED display.  
;  During a write, the LEDs should initilize to all '1' and then set to
;  '0', 1 bit at a time, with each bit indicating that another 1/8th of the
;  EEPROM array has been written.  The entire operation takes less than 1
;  second after applying power or resetting the part.  A 55H is output to
;  port B after the write has completed.  The data is initilized to FFH and
;  the address to 00H.  The address is incremented by 1 and the data
;  decremented by 1 after each byte is written, until the EE array is full.

; In this application note, the port B LEDs are used to display the array
;  content during a read operation.  A 1 second timer is called between each
;  byte so the data can be read on the LEDs.  Note that with the MTA85X1X
;  version (RB7 = SEEVDD),  RB7 must always = '1' when the EEPROM is being
;  accessed.  The EEPROM address pointer (used for a read current
;  address) will be reset if RB7 is set to a '0'.  Due to this, the RB7 pin
;  is always set to a '1' during a READ_CURRENT operation (even if the data
;  read is a '0' a one will be output to keep from resetting the EEPROM's
;  internal address).  MTA85X0X users may want to comment out the line which
;  sets the RB7 to a '1' during a READ_CURRENT operation, so that they will
;  receive the data read from the EEPROM on the LEDs without RB7 altered.

; NOTE: Code is not strictly I2C compattable on all timing parameters, but
; is designed to work with a Microchip 24LC01B or 24LC02B EEPROM, over full
; operating voltage.  Commented NOPs have been included which can be put in
; place to gain I2C compatibility.  Also note that '1's are driven onto the
; bus (not floated) by the master in most write circumstances, except where
; bus contnetion from the EEPROM slave could result.


;***************************************************************************
;***************************  Variable Listing  ****************************
;***************************************************************************
OK          EQU     01H
NO          EQU     00H
;INDIRECT   EQU     00H         ; Indirect Address Register
;RTCLOCK    EQU     01H         ; Real time counter clock
PC          EQU     02H         ; Program counter
STATUS      EQU     03H         ; Status register
;PD         EQU     03H         ; Power down bit in STATUS
PORTA       EQU     05H         ; Port A control register
PORTB       EQU     06H         ; Port B control register
I2C_PORT    EQU     05H         ; Port A control register, used for I2C

SCL         EQU     02H         ; EEPROM Clock, SCL (I/O bit 2)
SDA         EQU     03H         ; EEPROM Data,  SDA (I/O bit 3)
EE_OK       EQU     07H         ; Bit 7 in PC_OFFSET used as OK flag for EE
PC_OFFSET   EQU     07H         ; PC offset register (low order 4 bits), 
				;  value based on operating mode of EEPROM.
				;  Also, bit 7 used for EE_OK flag
EEADDR      EQU     08H         ; EEPROM Address
EEDATA      EQU     09H         ; EEPROM Data
EEBYTE      EQU     0AH         ; Byte sent to or received from
				; EEPROM (control, address, or data)
COUNTER     EQU     0BH         ; Bit counter for serial transfer

   ORG      1FFH
   GOTO     START

   ORG     0

;***************************************************************************
;***************************  EEPROM Subroutines  **************************
;***************************************************************************
; Communication for EEPROM based on I2C protocall, with Acknowledge.
;
; Byte_Write: Byte write routine
;       Inputs:  EEPROM Address   EEADDR
;                EEPROM Data      EEDATA
;       Outputs: Return 01 in W if OK, else return 00 in W
;
; Read_Current: Read EEPROM at address currently held by EE device. 
;       Inputs:  NONE
;       Outputs: EEPROM Data       EEDATA
;                Return 01 in W if OK, else return 00 in W
;
; Read_Random: Read EEPROM byte at supplied address
;       Inputs:  EEPROM Address    EEADDR
;       Outputs: EEPROM Data       EEDATA
;                Return 01 in W if OK, else return 00 in W
;
; Note: EEPROM subroutines will set bit 7 in PC_OFFSET register if the
;       EEPROM acknowledged OK, else that bit will be cleared.  This bit 
;       can be checked instead of refering to the value returned in W
;***************************************************************************

;********************** Set up EEPROM control bytes ************************
;***************************************************************************
READ_CURRENT
    MOVLW   B'10000100'   ; PC offset for read current addr.  EE_OK bit7='1'
    MOVWF   PC_OFFSET     ; Load PC offset
    GOTO    INIT_READ_CONTROL

WRITE_BYTE
    MOVLW   B'10000000'   ; PC offset for write byte.  EE_OK: bit7 = '1'
    GOTO    INIT_WRITE_CONTROL

READ_RANDOM
    MOVLW   B'10000011'   ; PC offset for read random.  EE_OK: bit7 = '1'

INIT_WRITE_CONTROL
    MOVWF   PC_OFFSET     ; Load PC offset register, value preset in W
    MOVLW   B'10100000'   ; Control byte with write bit, bit 0 = '0'
  
START_BIT
    BCF     I2C_PORT,SDA  ; Start bit, SDA and SCL preset to '1'


;******* Set up output data (control, address, or data) and counter ********
;***************************************************************************
PREP_TRANSFER_BYTE
    MOVWF   EEBYTE        ; Byte to transfer to EEPROM already in W
    MOVLW   B'00000011'   ; SDA and SCL set to output
    TRIS    I2C_PORT
    MOVLW   .8            ; Counter to transfer 8 bits
    MOVWF   COUNTER


;************  Clock out data (control, address, or data) byte  ************
;***************************************************************************
OUTPUT_BYTE
    BCF     I2C_PORT,SCL  ; Set clock low during data set-up
    RLF     EEBYTE        ; Rotate left, high order bit into carry bit
    BCF     I2C_PORT,SDA  ; Set data low, if rotated carry bit is
    SKPNC                 ;   a '1', then:
    BSF     I2C_PORT,SDA  ; reset data pin to a one, otherwise leave low
    BSF     I2C_PORT,SCL  ; clock data into EEPROM
    DECFSZ  COUNTER       ; Repeat until entire byte is sent
    GOTO    OUTPUT_BYTE


;**************************  Acknowkedge Check *****************************
;***************************************************************************
    MOVLW   B'00001011'   ; SDA = input, SCL = output
    SKPNC                 ; if SDA = 1 then tristate port to allow
    TRIS    I2C_PORT      ;   pullup to hold '1', avoiding bus contention
			  ;   if EEPROM acks in < 1us after clock goes low
    BCF     I2C_PORT,SCL  ; Set SCL low, 0.5us < ack valid < 3us
    TRIS    I2C_PORT      ; If SDA = '0' wait until SCL is low to set SDA to
			  ;   input. If done above, could have sent STOP bit
;   NOP                   ; May be necessary for SCL Tlow  at low voltage,
			  ;  also give resistor time to pull up bus if last 
			  ;  bit written = '0' and there is no ack from slave
    BSF     I2C_PORT,SCL  ; Raise SCL, EEPROM acknowledge still valid
    BTFSC   I2C_PORT,SDA  ; Check SDA for acknowledge (low)
    BCF     PC_OFFSET,EE_OK ; If SDA not low (no ack), set error flag
    BCF     I2C_PORT,SCL    ; Lower SCL, EEPROM release bus
    BTFSS   PC_OFFSET,EE_OK ; If no error continue, else stop bit
    GOTO    STOP_BIT


;*****  Set up program counter offset, based on EEPROM operating mode  *****
;***************************************************************************
    MOVF    PC_OFFSET,W
    ANDLW   B'00001111'
    ADDWF   PC
    GOTO    INIT_ADDRESS      ;PC offset=0, write control done, send address
    GOTO    INIT_WRITE_DATA   ;PC offset=1, write address done, send data
    GOTO    STOP_BIT          ;PC offset=2, write done, send stop bit
    GOTO    INIT_ADDRESS      ;PC offset=3, write control done, send address
    GOTO    INIT_READ_CONTROL ;PC offset=4, send read control
    GOTO    READ_BIT_COUNTER  ;PC offset=5, set counter and read byte
    GOTO    STOP_BIT          ;PC offset=6, random read done, send stop


;**********  Initalize EEPROM data (address, data, or control) bytes  ******
;***************************************************************************
INIT_ADDRESS
    INCF    PC_OFFSET    ; Increment PC offset to 2 (write) or to 4 (read)
    MOVF    EEADDR,W     ; Put EEPROM address in W, ready to send to EEPROM
    GOTO    PREP_TRANSFER_BYTE

INIT_WRITE_DATA
    INCF    PC_OFFSET    ; Increment PC offset to go to STOP_BIT next
    MOVF    EEDATA,W     ; Put EEPROM data in W, ready to send to EEPROM
    GOTO    PREP_TRANSFER_BYTE

INIT_READ_CONTROL
    BSF     I2C_PORT,SCL ; Raise SCL
    INCF    PC_OFFSET    ; Increment PC offset to go to READ_BIT_COUNTER next
    MOVLW   B'10100001'  ; Set up read control byte, ready to send to EEPROM
    GOTO    START_BIT    ;   bit 0 = '1' for read operation


;**************************  Read EEPROM data  *****************************
;***************************************************************************
READ_BIT_COUNTER
    MOVLW   .8           ; Set counter so 8 bits will be read into EEDATA
    MOVWF   COUNTER

READ_BYTE    
    BSF     I2C_PORT,SCL ; Raise SCL, SDA valid.  SDA still input from ack
    SETC                 ; Assume bit to be read = 1
    BTFSS   I2C_PORT,SDA ; Check if SDA = 1
    CLRC                 ; if SDA not = 1 then clear carry bit
    RLF     EEDATA       ; rotate carry bit (=SDA) into EEDATA;
    BCF     I2C_PORT,SCL ; Lower SCL
    DECFSZ  COUNTER      ; Decrement counter
    GOTO    READ_BYTE    ; Read next bit if not finished reading byte


;******************  Generate a STOP bit and RETURN  ***********************
;***************************************************************************
STOP_BIT
    BCF     I2C_PORT,SDA ; SDA=0, on TRIS, to prepare for transition to '1' 
    MOVLW   B'00000011'  ; SDA and SCL set to outputs, Bit0 and Bit1 ' input
    TRIS    I2C_PORT
    BSF     I2C_PORT,SCL ; SCL = 1 to prepare for STOP bit
;   NOP                  ; 4 NOPs neccessary for I2C spec Tsu:sto = 4.7us                  
    BSF     I2C_PORT,SDA ; Stop bit, SDA transition to '1' while SCL high
    
    BTFSS   PC_OFFSET,EE_OK ; Check for error
    RETLW   NO              ; if error, send back NO 
    RETLW   OK              ; if no error, send back OK

    ;Note: SDA and SCL still being driven by master, both set to outputs.
;****************************************************************************
;************************  End EEPROM Subroutines  **************************


;***************************************************************************
;*****************************  MAIN routine  ******************************
;***************************************************************************
START

;***************************  Initalize EEPROM  ****************************
;***************************************************************************
    MOVLW   B'00000011'    ; SDA and SCL = output, Bit0 and Bit1 = input
    TRIS    I2C_PORT
    BSF     I2C_PORT,SCL   ; Initalize SCL and SDA to '1'
    BSF     I2C_PORT,SDA

;--------------------  Cut from here up to get EEPROM subroutines, ---------
;--------- EEPROM initilization, and variable names.  Also note that -------
;-------- that a GOTO START in last address and ORG 0 are done above. ----------------


;*** VARIALBLE FOR MAIN ROUTINE THAT ARE NOT USED IN EEPROM SUBROUTINES ****
;*******  THESE VARIABLES CAN BE ELIMINATED IN CUSTOMERS APPLICATION  ******
;****************  WITHOUT EFFECTING THE EEPROM SUBROUTINES  ***************
COUNT       EQU     0CH ; REGISTER FOR # OF BYTES IN ARRAY WRITTEN OR READ
LEDS        EQU     0DH ; REGISTER TO HOLD DATA TO OUTPUT TO PORT
TIMER_OL    EQU     0EH ; REGISTER FOR OUTER  LOOP VARIABLE FOR TIMER
TIMER_ML    EQU     0FH ; REGISTER FOR MIDDLE LOOP VARIABLE FOR TIMER
TIMER_IL    EQU     10H ; REGISTER INNER  LOOP VARIABLE FOR TIMER
EE1KPARTIAL EQU     10H ; 1/8TH OF 1K ARRAY (128 BYTES TOTAL)
EE2KPARTIAL EQU     20H ; 1/8TH OF 2K ARRAY (256 BYTES TOTAL)
EE1KFULL    EQU     80H ; 128
EE2KFULL    EQU     00H ; WILL = FFH AFTER DECREMENT, EFFECTIVELY = 256
SIZE        EQU     00H ; BIT0 ON PORT WILL BE READ TO DETERMINE ARRAY SIZE
READORWRITE EQU     01H ; BIT1 ON PORT WILL BE READ TO DETERMINE IF USER 
			; WANTS TO READ OR WRITE THE ARRAY 
EEVDD       EQU     07H ; BIT FOR EEPROM VDD ON MTA85X1X DEVICES


;**********************  Read / Write  Set-Up  *****************************
;***************************************************************************

    MOVLW   000H
    TRIS    PORTB             ; PORTB = OUTPUT
    MOVWF   EEADDR            ; 1ST ADDRESS = 00H
    MOVLW   0FFH              ; INITILIZE DATA TO 'FF'
    MOVWF   LEDS              ; INITILIZE PORTB OUTPUT FOR % DONE INDICATOR
    MOVWF   PORTB             ; SET PORT B TO ALL '1's
    MOVWF   EEDATA            ; INITIALIZE DATA TO ALL '1'

    BTFSS   PORTA,READORWRITE ; IF RA1 = '1' THEN WRITE, RA1 = '0' THEN READ
    GOTO    SETSIZE_READ

SET_SIZE
    MOVLW   EE1KPARTIAL       ; INIT COUNTER TO 1/8 OF ARRAY SIZE, 1K
    BTFSC   PORTA,SIZE        ; IF RA0 = '0' THEN 1K EEPROM, IF '1' THEN 2K
    MOVLW   EE2KPARTIAL       ; INIT COUNTER TO 1/8 OF ARRAY SIZE, 2K
    MOVWF   COUNT             


;*****************************  Write Array  ********************************
;****************************************************************************
CALL_WRITE
    CALL    WRITE_BYTE
    CLRWDT                    ; DON'T TIME OUT DURING EE OPERATIONS
    BTFSS   PC_OFFSET,EE_OK   ; TEST EE_OK BIT DETERMINED DURING ACK CHECK
    GOTO    CALL_WRITE        ; IF NO ACCKNOWLEDGE, ASSUME PART IS STILL
    INCF    EEADDR            ; WRITING AND KEEP TRYING UNTIL GET THROUGH.
    DECF    EEDATA            ; IF PART DOES ACK, CHANGE DATA AND ADDRESS
    DECFSZ  COUNT             ; EVERY 1/8TH OF ARRAY CHANGE OUTPUT LEDs
    GOTO    CALL_WRITE      


;    *******  UPDATE LED OUTPUTS TO INDICATE ARRAY BEING WRITTEN *******
    BTFSS   LEDS,06H          ; IF BIT6 IS CLEAR BEFORE SHIFT THEN ARRAY IS
    GOTO    INFINITELOOP      ; FINISHED WRITING (EXCEPT LAST BYTE STILL NOT
			      ; DONE) DO NOT CLEAR BIT7 OR MTA85X1X PRODUCTS
			      ; WILL LOSE THEIR VDD AND NOT FINISH LAST BYTE 
    BCF     STATUS,0          ; SET CARRY BIT TO '0'
    RLF     LEDS              ; INCREMENT % DONE INDICATOR
    MOVF    LEDS,W
    MOVWF   PORTB
    GOTO    SET_SIZE

;*************************  Timer 1-second  *********************************
;****************************************************************************
TIMER
    MOVLW    .107 
    MOVWF    TIMER_OL
OUTTER_LOOP
    CLRWDT
    MOVLW    .25
    MOVWF    TIMER_ML
MIDDLE_LOOP
    MOVLW    .123
    MOVWF    TIMER_IL
INNER_LOOP
    DECFSZ   TIMER_IL
    GOTO     INNER_LOOP        
    DECFSZ   TIMER_ML
    GOTO     MIDDLE_LOOP
    DECFSZ   TIMER_OL
    GOTO     OUTTER_LOOP
    RETLW    00H
;***************************************************************************    


;****************************  Read Array  *********************************
;****************************************************************************
SETSIZE_READ
    MOVLW   EE1KFULL          ; INIT COUNTER TO ARRAY SIZE, 1K
    BTFSC   PORTA,SIZE        ; IF RA0 = '0' THEN 1K EEPROM, IF '1' THEN 2K
    MOVLW   EE2KFULL          ; INIT COUNTER TO ARRAY SIZE, 2K
    MOVWF   COUNT             
    
CALL_READ_RANDOM
    CALL    READ_RANDOM
    BTFSS   PC_OFFSET,EE_OK   ; TEST EE_OK BYTE DETERMINED DURING ACK CHECK
    GOTO    CALL_READ_RANDOM  ; IF NO ACCKNOWLEDGE, TRY AGAIN UNTIL RESET.
    INCF    EEADDR            
    MOVF    EEDATA,W
    MOVWF   PORTB
    CALL    TIMER             ; SLOW DOWN OUTPUT SO IT IS READABLE ON LEDs
    BSF     PORTB,EEVDD       ; Always set RB7 to '1' for MTA85X1X devices
    DECFSZ  COUNT             ;   May see flash on LED.  If rise time is
    GOTO    CALL_READ_RANDOM  ;   to slow, EEPROM may not be powered-up yet

; RESET SIZE_READ
    MOVLW   EE1KFULL          ; INIT COUNTER TO ARRAY SIZE, 1K
    BTFSC   PORTA,SIZE        ; IF RA0 = '0' THEN 1K EEPROM, IF '1' THEN 2K
    MOVLW   EE2KFULL          ; INIT COUNTER TO ARRAY SIZE, 2K
    MOVWF   COUNT             
    
CALL_READ_CURRENT
    CALL    READ_CURRENT
    BTFSS   PC_OFFSET,EE_OK   ; TEST EE_OK BYTE DETERMINED DURING ACK CHECK
    GOTO    CALL_READ_CURRENT ; IF NO ACKNOWLEDGE, TRY AGAIN UNTIL RESET.
    BSF     EEDATA,EEVDD       ; Always set RB7 to '1' for MTA85X1X devices
    MOVF    EEDATA,W          ;  During READ_CURRENT RB7 CANNOT go to a '0'
    MOVWF   PORTB             ;  or EEPROM internal address will be reset                                              
    CALL    TIMER             ; SLOW DOWN OUTPUT SO IT IS READABLE ON LEDs
    DECFSZ  COUNT              
    GOTO    CALL_READ_CURRENT    
;   NOTE: Data observed on port B will always show a '1' regardless of what
;         was read.  MTA85X1X pinout dedvices (RB7=SVDD) will reset the EEPROM
;         internal address pointer if RB7 = 0.  If the user has the MTA85X0X
;         pinout, the BSF EEDATA,EEVDD line above can be commented out. 

;**************************  End of Program Loop  **************************
;***************************************************************************
INFINITELOOP
    MOVLW   0AAH    ; NOTE: Bit7 = '1' to let EEPROM finish writing, MTA85X1X
    MOVWF   PORTB
    CLRWDT
    GOTO    INFINITELOOP

;****************************************************************************
;*******************************  End MAIN  *********************************
   END
